"""Module with functions for calculating the normal modes and frequencies of a given hessian. Eigenvalue/vector functions.
Also contains functions for calculating mechanical properties from NMA results. I.e. squared fluctuations, distance
fluctuations, etc."""
import numba as nb
import numpy as np


def modeCalc(hess, n_modes=200, eigen_method='eigsh', is3d=True, shift_invert=True, save_modes=True, save_mode_path='./'):
    """Calculate the 'n_modes' lowest frequency modes of the system by calculating the smallest eigenvalues and eigenvectors
    of the hessian matrix.

    :param hess: Sparse hessian matrix
    :param n_modes: Integer number of low-frequency modes to calculate.
    :param eigen_method: Choice of method for solving the eigenvalue problem.
    :returns:
    """
    from timeit import default_timer as timer
    NMA_start = timer()
    print('Calculating Normal Modes')

    n_dim = hess.shape[0]

    # if useMass:
    #     from scipy.sparse import diags
    #     if model == 'anm':
    #         masses = np.repeat(masses, 3)
    #
    #     mass = np.tile(masses, 60)
    #     print('mass variance: ', np.std(mass))
    #     print('mass mean: ', np.mean(mass))
    #     print('mass min: ', np.min(mass))
    #     print('mass max: ', np.max(mass))
    #     Mass = diags(mass)
    # else:


    # from scipy.sparse import identity
    # Mass = identity(n_dim)

    if eigen_method == 'eigsh':
        from scipy.sparse.linalg import eigsh
        if shift_invert:
            print('Using shift-invert for increased performance with increased memory usage.')
            evals, evecs = eigsh(hess, k=n_modes, sigma=1e-10, which='LA')
        else:
            evals, evecs = eigsh(hess, k=n_modes, which='SA')
            evals = evals[6:]
            evecs = evecs[:, 6:]
    elif eigen_method == 'lobpcg':
        from scipy.sparse.linalg import lobpcg

        epredict = np.random.rand(n_dim, n_modes + 6)
        evals, evecs = lobpcg(hess, epredict, largest=False, tol=0, maxiter=n_dim)
        evals = evals[6:]
        evecs = evecs[:, 6:]
    elif eigen_method == 'lobcuda':
        import cupy as cp
        from cupyx.scipy.sparse.linalg import lobpcg as clobpcg
        from cupyx.scipy.sparse.linalg import LinearOperator, spilu

        sparse_gpu = cp.sparse.csr_matrix(hess.astype(cp.float32))
        epredict = cp.random.rand(n_dim, n_modes + 6, dtype=cp.float32)

        lu = spilu(sparse_gpu, fill_factor=50)  # LU decomposition
        M = LinearOperator(hess.shape, lu.solve)
        print('gpu eigen')

        evals, evecs = clobpcg(sparse_gpu, epredict, M=M, largest=False, tol=0, verbosityLevel=0)
        if is3d:
            evals = cp.asnumpy(evals[6:])
            evecs = cp.asnumpy(evecs[:, 6:])
        else:
            evals = cp.asnumpy(evals[1:])
            evecs = cp.asnumpy(evecs[:, 1:])

    NMA_time = timer() - NMA_start
    print('NMA time: ', NMA_time)

    if save_modes:
        file = save_mode_path + 'modes.npz'
        print('Saving NMA results in' + file)
        np.savez_compressed(file, eigen_vals=evals, eigen_vecs = evecs)

    return evals, evecs


def calcCovMat(evals, evecs, n_modes, coords, fluct_cutoff, is3d=True):
    """Calculates a sparse covariance matrix from the low frequency modes.

    :param evals:
    :param evecs:
    :param n_modes:
    :param coords:
    :param fluct_cutoff:
    :param is3d:
    :return:
    """
    from scipy import sparse
    from sklearn.neighbors import BallTree, radius_neighbors_graph

    print('Calculating sparse covariance matrix')

    tree = BallTree(coords)
    adj = radius_neighbors_graph(tree, fluct_cutoff, mode='connectivity', n_jobs=-1).tocoo()
    adj.setdiag(1)
    row, col = adj.row, adj.col

    if is3d:
        data = con_c(evals, evecs, row, col)
        covariance = sparse.coo_matrix((data, (row, col)), shape=adj.shape)
    else:
        data = gCon_c(evals, evecs, row, col)
        covariance = sparse.coo_matrix((data, (row, col)), shape=adj.shape)

    return covariance


def calcDistFlucts(evals, evecs, coords, fluct_cutoff=7.5, is3d=True, **kwargs):
    """Calculates a sparse covariance matrix from the low frequency modes.

    :param evals:
    :param evecs:
    :param n_modes:
    :param coords:
    :param fluct_cutoff:
    :param is3d:
    :return:
    """
    from scipy import sparse

    n_modes = evals.shape[0]
    covariance = calcCovMat(evals, evecs, n_modes, coords, fluct_cutoff, is3d)
    c_diag = covariance.diagonal()
    row, col, c_data = (covariance.row, covariance.col, covariance.data)

    print('Calculating sparse distance fluctuation matrix from covariance matrix')
    fluct_data = distFluctFromCov(c_diag, c_data, row, col)

    dist_flucts = sparse.coo_matrix((fluct_data, (row, col)), shape=covariance.shape)

    return dist_flucts


@nb.njit()
def con_c(evals, evecs, row, col):
    """

    :param evals:
    :param evecs:
    :param row:
    :param col:
    :return:
    """
    data = []
    for k in range(row.shape[0]):
        i, j = (row[k], col[k])
        data.append(cov(evals, evecs, i, j))
    return np.array(data)


#@nb.njit()
def gCon_c(evals, evecs, row, col):
    """

    :param evals:
    :param evecs:
    :param row:
    :param col:
    :return:
    """
    data = []
    for k in range(row.shape[0]):
        i, j = (row[k], col[k])
        data.append(gCov(evals, evecs, i, j))
    return np.array(data)


@nb.njit(parallel=False)
def cov(evals, evecs, i, j):
    """

    :param evals:
    :param evecs:
    :param i:
    :param j:
    :return:
    """
    # Calculates the covariance between two residues in ANM. Takes the trace of block so no anisotropy info.
    # Commented section would compute normalized covariances
    n_e = evals.shape[0]
    # n_d = evecs.shape[1]
    tr1 = 0
    # tr2 = 0
    # tr3 = 0
    for n in range(n_e):
        l = evals[n]
        tr1 += 1 / l * (evecs[3 * i, n] * evecs[3 * j, n] + evecs[3 * i + 1, n] * evecs[3 * j + 1, n] + evecs[
            3 * i + 2, n] * evecs[3 * j + 2, n])
        # tr2 += 1 / l * (evecs[3 * i, n] * evecs[3 * i, n] + evecs[3 * i + 1, n] * evecs[3 * i + 1, n] + evecs[
        #     3 * i + 2, n] * evecs[3 * i + 2, n])
        # tr3 += 1 / l * (evecs[3 * j, n] * evecs[3 * j, n] + evecs[3 * j + 1, n] * evecs[3 * j + 1, n] + evecs[
        #     3 * j + 2, n] * evecs[3 * j + 2, n])
    cov = tr1  # / np.sqrt(tr2 * tr3)
    return cov


@nb.njit(parallel=False)
def gCov(evals, evecs, i, j):
    """

    :param evals:
    :param evecs:
    :param i:
    :param j:
    :return:
    """
    # Calculates the covariance between two residues in GNM
    n_e = evals.shape[0]
    c = 0
    for n in range(n_e):
        l = evals[n]
        c += 1 / l * (evecs[i, n] * evecs[j, n])
    return c


@nb.njit()
def distFluctFromCov(c_diag, c_data, row, col):
    """

    :param c_diag:
    :param c_data:
    :param row:
    :param col:
    :return:
    """
    d_data = []
    for k in range(row.shape[0]):
        i, j = (row[k], col[k])
        d = np.abs(c_diag[i] + c_diag[j] - 2 * c_data[k])
        d_data.append(d)
    return np.array(d_data)


def fluctPlot(d, title, pdb):
    """

    :param d:
    :param title:
    :param pdb:
    """
    import matplotlib.pyplot as plt
    print('Plotting Fluctuation Histogram')
    import matplotlib
    fig, ax = plt.subplots(1, 1, figsize=(10, 5))
    font = {'family': 'sans-serif',
            'weight': 'normal',
            'size': 11}
    matplotlib.rc('font', **font)
    print('Plotting')
    ax.set_ylabel('Density', fontsize=12)
    ax.set_xlabel('$A^{2}$', fontsize=12)
    ax.tick_params(axis='y', labelsize=8)
    ax.tick_params(axis='x', labelsize=8)
    ax.legend()
    fig.suptitle(
        'Histogram Of Pairwise Fluctuations: ' + title + ' (' + pdb + ')', fontsize=12)

    ax.hist(d.data, bins='fd', histtype='stepfilled', density=True)
    fig.tight_layout()
    plt.show()


# This function calls the relevant function from bfactorfit.py
def fitCompareBfactors(evals, evecs, bfactors, pdb, is3d=True, fit_modes=True, plot_modes=True, force_ico=False,
                       ico_tol=0.002, is_ico=True, save_bfactors=True, save_bfactors_path='./'):
    """

    :param is_ico:
    :param evals:
    :param evecs:
    :param bfactors:
    :param pdb:
    :param is3d:
    :param fit_modes:
    :param plot_modes:
    :param force_ico:
    :param ico_tol:
    :return:
    """
    from .bfactorfit import fitPlotBfactors
    from timeit import default_timer as timer
    bfactor_start = timer()
    cc, gamma,n_modes = fitPlotBfactors(evals, evecs, bfactors, pdb, is3d=is3d, fitModes=fit_modes, plotModes=plot_modes,
                                        forceIco=force_ico, icotol=ico_tol, isIco=is_ico, save_path=save_bfactors_path)

    bfactor_time = timer() - bfactor_start
    print('bfactor fitting time: ', bfactor_time)
    r_evals = evals[:n_modes]*gamma
    r_evecs = evecs[:, :n_modes]
    return r_evals, r_evecs, cc, gamma, n_modes


"""Module with functions for identifying quasi-rigid clusters given a distance fluctuation matrix, either calculated using
ENM or provided."""

import numpy as np


def findQuasiRigidClusters(pdb, dist_flucts, n_range=None, cluster_start=4, cluster_stop=100, cluster_step=1,
                           cluster_method='discretize', return_type='final',
                           score_method='median', save_results=True, save_results_path='./', clust_mask=None, clust_filter_method='embedding', **kwargs):
    """Uses spectral clustering to split the residues into clusters with minimal internal distance fluctuations.

    :param str pdb:
    :param dist_flucts:
    :param list n_range:
    :param str cluster_method:
    :param str score_method:
    :param bool save_results:
    :param str dir:
    :return:
    """
    from timeit import default_timer as timer
    QRC_start = timer()

    if n_range is None:
        n_range = np.arange(cluster_start, cluster_stop, cluster_step)

    if clust_mask is None:
        clust_mask = np.full(dist_flucts.shape[0], True)

    if clust_filter_method == 'flucts' and clust_mask is not None:
        dist_flucts = filterFlucts(dist_flucts, clust_mask)
        mask = None
    else:
        mask = clust_mask

    sims = fluctToSims(dist_flucts)

    n_vecs = n_range.max()

    embedding = calcEmbedding(sims, n_vecs)

    labels, scores, numtypes, full_scores = cluster_embedding(n_range, embedding, method=cluster_method,
                                                              score_method=score_method, clust_mask=mask)

    QRC_time = timer() - QRC_start
    print('QRC time: ', QRC_time)

    from .clustering_util import plotScores
    plotScores(pdb, n_range, scores, numtypes, save_path=save_results_path)

    ind = np.argmax(scores)
    final_cluster_num = n_range[ind]
    final_clusters = labels[ind]
    final_score = scores[ind]
    final_numtypes = numtypes[ind]
    final_full_score = full_scores[ind]
    np.savez_compressed(save_results_path + pdb + '_' + return_type + '_results_full', labels=labels, score=scores,
                        nc_range=n_range, cluster_method=cluster_method, numtypes=numtypes, full_scores=full_scores, mask=clust_mask)
    if return_type == 'final':
        if save_results:
            np.savez_compressed(save_results_path + pdb + '_' + return_type + '_results', labels=final_clusters,
                                score=final_score,
                                nc=final_cluster_num, cluster_method=cluster_method, final_full_score=final_full_score)
        return final_clusters, final_score, final_full_score
    elif return_type == 'full':
        if save_results:
            np.savez_compressed(save_results_path + pdb + '_' + return_type + '_results', labels=labels, score=scores,
                                nc_range=n_range, cluster_method=cluster_method, numtypes=numtypes,
                                full_scores=full_scores)
        return labels, scores, numtypes, full_scores
    else:
        return final_clusters


def fluctToSims(d):
    """Transforms a distance fluctuation matrix into a similarity matrix

    :param d: Sparse distance fluctuation matrix
    :return: Sparse similarity matrix
    """
    from scipy import sparse
    d_bar = np.mean(np.sqrt(d.data))
    # print('RMS distance fluctuations: ', d_bar)
    sigma = 1 / (2 * d_bar ** 2)
    data = d.data
    data = np.exp(-sigma * data ** 2)
    sims = sparse.coo_matrix((data, (d.row, d.col)), shape=d.shape)
    sims.eliminate_zeros()
    return sims

import numpy as np
def filterChains(calphas, clust_ignore_chains):
    clust_mask = np.full(len(calphas), True)
    for chain in clust_ignore_chains:
        mask = np.logical_not(calphas.chain_id == chain)
        clust_mask = clust_mask & mask
    print(clust_mask)
    return clust_mask

def filterEmbedding(vecs, mask):
    #print('Filtering some chains out of clustering process')
    indices = mask.nonzero()[0]
    vecs_new = vecs[indices, :].copy()
    return vecs_new

def filterFlucts(flucts, mask):
    #print('Filtering some chains out of dist_fluct matrix')
    indices = mask.nonzero()[0]
    flucts = flucts.tocsr()
    flucts_new = flucts[:, indices][indices, :].copy()
    return flucts_new.tocoo()


def calcEmbedding(sims, n_vecs):
    """

    :param sims:
    :param n_vecs:
    :return:
    """
    from sklearn.manifold import spectral_embedding
    print('Performing Spectral Embedding')


    X_transformed = spectral_embedding(sims, n_components=n_vecs, drop_first=False, eigen_solver='arpack',
                                       norm_laplacian=True)

    return X_transformed


def cluster_embedding(n_range, maps, method='discretize', score_method='median', clust_mask=None):
    """

    :param n_range:
    :param maps:
    :param method:
    :param score_method:
    :return:
    """



    print('Clustering Embedded Points')
    print(f'Method: {method}')

    from sklearn.preprocessing import normalize
    if method == 'kmeans':
        from sklearn.cluster import k_means
    # from sklearn.cluster import MiniBatchKMeans, AgglomerativeClustering
    # from sklearn.metrics import pairwise_distances
    # from sklearn_extra.cluster import KMedoids

    # from sklearn.metrics import silhouette_score
    # from sklearn.metrics import davies_bouldin_score
    from .clustering_util import median_score, cluster_types, calcCentroids, calcCosCentroids

    if clust_mask is not None:
        maps = filterEmbedding(maps, clust_mask)

    labels = []
    fullScores = []
    scores = []
    numtypes = []

    for n in range(len(n_range)):
        n_clusters = n_range[n]
        emb = maps[:, :n_clusters].copy()

        normalize(emb, copy=False)

        # print('Clusters: ' + str(n_clusters))

        if method == 'discretize':
            label = discretize(emb)
            centroids = calcCosCentroids(emb, label, n_clusters)
        elif method == 'kmeans':
            centroids, label, inert, n_iter = k_means(emb, n_clusters=n_clusters, return_n_iter=True)
        else:
            raise Exception('Method should be kmeans or discretize.')

        labels.append(label)

        med_score, full_score = median_score(emb, centroids, score_method)
        # print('Score: ', testScore)
        fullScores.append(full_score)
        scores.append(med_score)

        var, ntypes = cluster_types(label)
        numtypes.append(ntypes)

    return labels, scores, numtypes, fullScores


# Adapted from sklearn. May need to change later.
def discretize(
        vectors, *, copy=False, max_svd_restarts=300, n_iter_max=2000, random_state=None
):
    """Search for a partition matrix which is closest to the eigenvector embedding.
    This implementation was proposed in [1]_.
 
    :param vectors: array-like of shape (n_samples, n_clusters)
        The embedding space of the samples.
    :param copy: bool, default=True
        Whether to copy vectors, or perform in-place normalization.
    :param max_svd_restarts: int, default=30
        Maximum number of attempts to restart SVD if convergence fails
    :param n_iter_max: int, default=30
        Maximum number of iterations to attempt in rotation and partition
        matrix search if machine precision convergence is not reached
    :param random_state: int, RandomState instance, default=None
        Determines random number generation for rotation matrix initialization.
        Use an int to make the randomness deterministic.
        See :term:`Glossary <random_state>`.
    Returns
    -------
    labels: array of integers, shape: n_samples
        The labels of the clusters.
    References
    ----------
    .. [1] `Multiclass spectral clustering, 2003
           Stella X. Yu, Jianbo Shi
           <https://www1.icsi.berkeley.edu/~stellayu/publication/doc/2003kwayICCV.pdf>`_
    Notes
    -----
    The eigenvector embedding is used to iteratively search for the
    closest discrete partition.  First, the eigenvector embedding is
    normalized to the space of partition matrices. An optimal discrete
    partition matrix closest to this normalized embedding multiplied by
    an initial rotation is calculated.  Fixing this discrete partition
    matrix, an optimal rotation matrix is calculated.  These two
    calculations are performed until convergence.  The discrete partition
    matrix is returned as the clustering solution.  Used in spectral
    clustering, this method tends to be faster and more robust to random
    initialization than k-means.
    """

    from scipy.sparse import csc_matrix
    from scipy.linalg import LinAlgError
    from sklearn.utils import check_random_state, as_float_array

    random_state = check_random_state(random_state)

    vectors = as_float_array(vectors, copy=copy)

    eps = np.finfo(float).eps
    n_samples, n_components = vectors.shape

    # Normalize the eigenvectors to an equal length of a vector of ones.
    # Reorient the eigenvectors to point in the negative direction with respect
    # to the first element.  This may have to do with constraining the
    # eigenvectors to lie in a specific quadrant to make the discretization
    # search easier.
    norm_ones = np.sqrt(n_samples)
    for i in range(vectors.shape[1]):
        vectors[:, i] = (vectors[:, i] / np.linalg.norm(vectors[:, i])) * norm_ones
        if vectors[0, i] != 0:
            vectors[:, i] = -1 * vectors[:, i] * np.sign(vectors[0, i])

    # Normalize the rows of the eigenvectors.  Samples should lie on the unit
    # hypersphere centered at the origin.  This transforms the samples in the
    # embedding space to the space of partition matrices.
    vectors = vectors / np.sqrt((vectors ** 2).sum(axis=1))[:, np.newaxis]

    svd_restarts = 0
    has_converged = False

    # If there is an exception we try to randomize and rerun SVD again
    # do this max_svd_restarts times.
    while (svd_restarts < max_svd_restarts) and not has_converged:

        # Initialize first column of rotation matrix with a row of the
        # eigenvectors
        rotation = np.zeros((n_components, n_components))
        rotation[:, 0] = vectors[random_state.randint(n_samples), :].T

        # To initialize the rest of the rotation matrix, find the rows
        # of the eigenvectors that are as orthogonal to each other as
        # possible
        c = np.zeros(n_samples)
        for j in range(1, n_components):
            # Accumulate c to ensure row is as orthogonal as possible to
            # previous picks as well as current one
            c += np.abs(np.dot(vectors, rotation[:, j - 1]))
            rotation[:, j] = vectors[c.argmin(), :].T

        last_objective_value = 0.0
        n_iter = 0

        while not has_converged:
            n_iter += 1

            t_discrete = np.dot(vectors, rotation)

            labels = t_discrete.argmax(axis=1)
            vectors_discrete = csc_matrix(
                (np.ones(len(labels)), (np.arange(0, n_samples), labels)),
                shape=(n_samples, n_components),
            )

            t_svd = vectors_discrete.T * vectors

            try:
                U, S, Vh = np.linalg.svd(t_svd)
            except LinAlgError:
                svd_restarts += 1
                print("SVD did not converge, randomizing and trying again")
                break

            ncut_value = 2.0 * (n_samples - S.sum())
            if (abs(ncut_value - last_objective_value) < eps) or (n_iter > n_iter_max):
                has_converged = True
            else:
                # otherwise calculate rotation and continue
                last_objective_value = ncut_value
                rotation = np.dot(Vh.T, U.T)

    if not has_converged:
        raise LinAlgError("SVD did not converge")
    return labels


# Also adapted from sklearn, similar method?
# def cluster_qr(vectors):
#     from scipy.linalg import svd, qr
#     """Find the discrete partition closest to the eigenvector embedding.
#         This implementation was proposed in [1]_.
#     .. versionadded:: 1.1
#         Parameters
#         ----------
#         vectors : array-like, shape: (n_samples, n_clusters)
#             The embedding space of the samples.
#         Returns
#         -------
#         labels : array of integers, shape: n_samples
#             The cluster labels of vectors.
#         References
#         ----------
#         .. [1] :doi:`Simple, direct, and efficient multi-way spectral clustering, 2019
#             Anil Damle, Victor Minden, Lexing Ying
#             <10.1093/imaiai/iay008>`
#     """
#
#     k = vectors.shape[1]
#     _, _, piv = qr(vectors.T, pivoting=True)
#     ut, _, v = svd(vectors[piv[:k], :].T)
#     vectors = abs(np.dot(vectors, np.dot(ut, v.conj())))
#     return vectors.argmax(axis=1)



"""Module with functions for downloading and dealing with PDB/PDBx files."""


def getCapsid(pdb, save_pdb_path='./', pdbx=False, local=False, save_full_pdb=False, assembly_id=1, chains=None, chains_clust=None, is_ico=False):
    """Downloads and opens molecular data from a PDB entry or loads data from a local file.

    :param pdb: PDB id of entry to download. Can also be the name of a local file
    :param save_pdb_path: Target directory where download will be placed
    :param pdbx: Whether the target structure should be acquired in pdbx/mmcif format
    :param local: Whether to instead load a local file
    :param save_full_pdb: Whether to save a copy of the complete assembly as pdb/pdbx. Useful for some visualizations.
    :param chains: List of chains from the entry to include in the ENM model. Not implemented
    """


    if isinstance(pdbx, str):
        pdbx = pdbx == 'true' or pdbx == 'True'
    if isinstance(local, str):
        local = local == 'true' or local == 'True'
    if isinstance(save_full_pdb, str):
        save_full_pdb = save_full_pdb == 'true' or save_full_pdb == 'True'

    if local:
        filename = save_pdb_path + pdb
    else:
        filename = downloadPDB(pdb, save_pdb_path, pdbx)

    if pdbx:
        capsid, calphas, asym, coords, bfactors, chain_starts, title = loadPDBx(filename, pdb, save_full_pdb, assembly_id=assembly_id)
    else:
        try:
            capsid, calphas, asym, coords, bfactors, chain_starts, title = loadPDB(filename, pdb, save_full_pdb, assembly_id=assembly_id)
        except:
            print('no .pdb file found. Checking for pdbx/mmcif')
            print('Add pdbx=True if you know you will be fetching a pdbx/mmcif file')
            capsid, calphas, asym, coords, bfactors, chain_starts, title = loadPDBx(filename, pdb, save_full_pdb, assembly_id=assembly_id)

    n_res = len(calphas)
    print(f'# of residues: {n_res}')

    return capsid, calphas, asym, coords, bfactors, chain_starts, title


def downloadPDB(pdb, dir='.', pdbx=False):
    """Downloads pdb and returns the filename

    :param pdb: PDB id of entry to download. Can also be the name of a local file
    :param dir: Target directory where download will be placed
    :param pdbx: Whether the target structure should be acquired in pdbx/mmcif format
    """
    from biotite.database.rcsb import fetch
    if pdbx:
        filename = fetch(pdb, target_path=dir, format='pdbx', overwrite=True, verbose=True)
    else:
        try:
            filename = fetch(pdb, target_path=dir, format='pdb', overwrite=True, verbose=True)
        except:
            print('no .pdb file found for this PDB id. Checking for pdbx/mmcif')
            print('Add pdbx=True if you know you will be fetching a pdbx/mmcif file')
            filename = fetch(pdb, target_path=dir, format='pdbx', overwrite=True, verbose=True)
    return filename


def loadPDBx(filename, pdb, save, assembly_id=1):
    """Loads PDBx data from a file

    :param filename: Name of local file
    :param pdb: PDB id of entry
    :param save: Whether to save a copy of the complete assembly as pdb/pdbx.
        """
    import biotite.structure as struc
    import biotite.structure.io.pdbx as pdbx
    import biotite.structure.io as strucio
    import numpy as np

    pdbx_file = pdbx.CIFFile.read(filename)
    if assembly_id is not None:
        capsid = pdbx.get_assembly(pdbx_file, assembly_id=str(assembly_id), model=1, extra_fields=['b_factor'])
        asym = pdbx.get_structure(pdbx_file, model=1, extra_fields=['b_factor'])
    else:
        asym = pdbx.get_structure(pdbx_file, model=1, extra_fields=['b_factor'])
        capsid = asym

    capsid = capsid[struc.filter_amino_acids(capsid)]
    title = pdb  # pdbx_file.get_category('pdbx_database_related')['details']

    # capsid = capsid[capsid.filter_amino_acids()].copy()

    capsid = capsid[struc.filter_amino_acids(capsid)]
    asym = asym[struc.filter_amino_acids(asym)]
    print("Number of protein chains in full structure:", struc.get_chain_count(capsid))
    print("Number of protein chains in asymmetric unit:", struc.get_chain_count(asym))

    calphas = capsid[capsid.atom_name == 'CA']
    coords = calphas.coord

    if save:
        print('Saving complete capsid PDBx')
        strucio.save_structure(pdb + '_capsid.pdbx', capsid)

    chain_starts = struc.get_chain_starts(calphas)
    chain_starts = np.append(chain_starts, len(calphas))

    return capsid, calphas, asym, coords, calphas.b_factor, chain_starts, title

def loadPDB(filename, pdb_id, save, assembly_id=1):
    """Loads PDBx data from a file

    :param filename: Name of local file
    :param pdb_id: PDB id of entry
    :param save: Whether to save a copy of the complete assembly as pdb/pdbx.
        """
    import biotite.structure as struc
    import biotite.structure.io.pdb as pdb
    import biotite.structure.io as strucio
    import numpy as np

    pdb_file = pdb.PDBFile.read(filename)
    if assembly_id is not None:
        capsid = pdb.get_assembly(pdb_file, assembly_id=str(assembly_id), model=1, extra_fields=['b_factor'])
        asym = pdb.get_structure(pdb_file, model=1, extra_fields=['b_factor'])
    else:
        asym = pdb.get_structure(pdb_file, model=1, extra_fields=['b_factor'])
        capsid = asym

    title = pdb_id  # pdbx_file.get_category('pdbx_database_related')['details']

    # capsid = capsid[capsid.filter_amino_acids()].copy()

    capsid = capsid[struc.filter_amino_acids(capsid)]
    asym = asym[struc.filter_amino_acids(asym)]
    print("Number of protein chains in full structure:", struc.get_chain_count(capsid))
    print("Number of protein chains in asymmetric unit:", struc.get_chain_count(asym))

    calphas = capsid[capsid.atom_name == 'CA']
    coords = calphas.coord

    if save:
        print('Saving complete capsid PDB')
        strucio.save_structure(pdb_id + '_capsid.pdb', capsid, hybrid36=True)

    chain_starts = struc.get_chain_starts(calphas)
    chain_starts = np.append(chain_starts, len(calphas))

    return capsid, calphas, asym, coords, calphas.b_factor, chain_starts, title


# def loadPDB(filename, pdb, save):
#     """Loads PDBx data from a file
#
#     :param filename: Name of local file
#     :param pdb: PDB id of entry
#     :param save: Whether to save a copy of the complete assembly as pdb/pdbx
#     """
#     from prody import parsePDB, writePDB
#
#     capsid, header = parsePDB(filename, header=True, biomol=True, secondary=True, extend_biomol=True)
#     asym_unit = parsePDB(filename)  # Maybe need a check to see if assembly
#
#
#     ENM_capsid = capsid.select('protein').copy()
#     calphas = ENM_capsid.select('calpha')
#
#     ENM_capsid_asym = asym_unit.select('protein').copy()
#     calphas_asym = ENM_capsid_asym.select('calpha').copy()
#
#     chain_starts = getProdyChainStarts(calphas_asym)
#
#     print('Number Of Residues: ', calphas.getCoords().shape[0])
#
#     if save:
#         print('Writing complete capsid PDB')
#         #print('Number Of Saved Residues: ', ENM_capsid.numResidues())
#         writePDB(pdb + '_capsid.pdb', ENM_capsid, hybrid36=True)
#
#     if 'title' in header:
#         title = header['title']
#     else:
#         title = pdb
#
#     return ENM_capsid, calphas, calphas.getCoords(), calphas.getBetas(), chain_starts, title


# def getProdyChainStarts(calphas_asym, n_units=60):
#     """
#
#     :param calphas_asym:
#     :param n_units:
#     :return:
#     """
#     import numpy as np
#     n_asym = calphas_asym.numAtoms()
#     n_chains = calphas_asym.numChains()
#     chains = calphas_asym.getChids()
#     chaindiff = np.where(chains[:-1] != chains[1:])[0]
#     chain_starts = []
#     for i in range(n_units):
#         start = n_asym * i
#         chain_starts.append(start)
#         for c in chaindiff:
#             chain_starts.append(start + c)
#
#     chain_starts.append(n_asym * n_units)
#
#     return np.array(chain_starts)

# def buildMassesCoords(atoms):
#     print(atoms[0])
#     bfs = []
#     masses = []
#     prot_ids = []
#
#
#     for pn, chain in enumerate(atoms.iterChains()):
#         print('Chain Id ' + pn + ':' + chain.getChid())
#         prot_ids.append(pn)
#         for res in chain.iterResidues():
#             mass = np.sum(res.getMasses())
#             masses.append(mass)
#
#             bfactor = np.mean(res.getBetas())
#             bfs.append(bfactor)
#
#     return np.asarray(bfs), np.asarray(masses), np.asarray(pn)



# Colormap generation taken from stackexchange
# https://stackoverflow.com/questions/42697933/colormap-with-maximum-distinguishable-colours
def generate_colormap(number_of_distinct_colors: int = 80):
    import math

    import numpy as np
    from matplotlib.colors import ListedColormap
    from matplotlib.cm import hsv

    if number_of_distinct_colors == 0:
        number_of_distinct_colors = 80

    number_of_shades = 7
    if number_of_distinct_colors < (number_of_shades - 1):
        number_of_shades = number_of_distinct_colors - 2

    number_of_distinct_colors_with_multiply_of_shades = int(
        math.ceil(number_of_distinct_colors / number_of_shades) * number_of_shades)

    # Create an array with uniformly drawn floats taken from <0, 1) partition
    linearly_distributed_nums = np.arange(
        number_of_distinct_colors_with_multiply_of_shades) / number_of_distinct_colors_with_multiply_of_shades

    # We are going to reorganise monotonically growing numbers in such way that there will be single array with saw-like pattern
    #     but each saw tooth is slightly higher than the one before
    # First divide linearly_distributed_nums into number_of_shades sub-arrays containing linearly distributed numbers
    arr_by_shade_rows = linearly_distributed_nums.reshape(number_of_shades,
                                                          number_of_distinct_colors_with_multiply_of_shades // number_of_shades)

    # Transpose the above matrix (columns become rows) - as a result each row contains saw tooth with values slightly higher than row above
    arr_by_shade_columns = arr_by_shade_rows.T

    # Keep number of saw teeth for later
    number_of_partitions = arr_by_shade_columns.shape[0]

    # Flatten the above matrix - join each row into single array
    nums_distributed_like_rising_saw = arr_by_shade_columns.reshape(-1)

    # HSV colour map is cyclic (https://matplotlib.org/tutorials/colors/colormaps.html#cyclic), we'll use this property
    initial_cm = hsv(nums_distributed_like_rising_saw)

    lower_partitions_half = number_of_partitions // 2
    upper_partitions_half = number_of_partitions - lower_partitions_half

    # Modify lower half in such way that colours towards beginning of partition are darker
    # First colours are affected more, colours closer to the middle are affected less
    lower_half = lower_partitions_half * number_of_shades
    for i in range(3):
        initial_cm[0:lower_half, i] *= np.arange(0.2, 1, 0.8 / lower_half)

    # Modify second half in such way that colours towards end of partition are less intense and brighter
    # Colours closer to the middle are affected less, colours closer to the end are affected more
    for i in range(3):
        for j in range(upper_partitions_half):
            modifier = np.ones(number_of_shades) - initial_cm[lower_half + j * number_of_shades: lower_half + (
                        j + 1) * number_of_shades, i]
            modifier = j * modifier / upper_partitions_half
            initial_cm[lower_half + j * number_of_shades: lower_half + (j + 1) * number_of_shades, i] += modifier

    return ListedColormap(initial_cm)

from chimerax.core.commands import *
from chimerax.std_commands import *
from chimerax.core.commands import all_objects
from chimerax.core.commands import run
from chimerax.atomic.molsurf import MolecularSurface
from chimerax.atomic import *
import numpy as np
import matplotlib as mpl
# Colormap taken from stackexchange

# from chimerax.core.commands import measure


from chimerax.std_commands import split
from numpy.linalg import norm
import sys
print(sys.argv)
color_dir, save_dir, pdb_dir, remote, pdb, rwb_scale = (sys.argv[1], sys.argv[2], sys.argv[3], sys.argv[4], sys.argv[5], sys.argv[6])

if remote=='True':
    run(session, f'open {pdb}')
    run(session, 'sym #1 assembly 1 copies True')
    run(session, 'close #1')
else:
    run(session, 'open ' + '\'' + pdb_dir + '\'')

# center of mass measurement from ChimeraX tutorial
run(session, 'sel protein')
run(session, 'del ~sel')
run(session, 'sel clear')
atoms = all_objects(session).atoms  # getting atom list
coords = atoms.scene_coords  # getting atom coords
modelCenter = coords.mean(axis=0)  # calculate center of mass
print("modelCenter", modelCenter)

if modelCenter.any:
    print('Aligning Models')
    modelX, modelY, modelZ = modelCenter
    run(session, 'move x ' + str(-1 * modelX) + ' coordinateSystem #2 models #2')  # adjust x coordinates
    run(session, 'move y ' + str(-1 * modelY) + ' coordinateSystem #2 models #2')  # adjust y coordinates
    run(session, 'move z ' + str(-1 * modelZ) + ' coordinateSystem #2 models #2')  # adjust z coordinates
else:
    print('Models Already Aligned')

coords = atoms.scene_coords
radius = norm(coords, axis=1).max()  # calculate the norms of the x coordinates, and then choose the maximum value
print("radius_est: ", radius)

rgba = np.load(color_dir)
# import matplotlib as mpl
# norm = mpl.colors.Normalize(vmin=np.min(labels), vmax=np.max(labels))
#
# if rwb_scale=='True':
#     import matplotlib.pyplot as plt
#     cmap = plt.get_cmap('coolwarm_r')
#     rgba = cmap(norm(labels))*255
# else:
#     cmap = generate_colormap(int(np.max(labels)))
#     rgba = cmap(norm(labels))*255

residues = atoms.unique_residues
cx_nr = len(residues)
enm_nr = rgba.shape[0]

print('# of residues:', cx_nr)
print('# of ENM residues:', enm_nr)

for i in range(cx_nr):
    res = residues[i]
    res.ribbon_color = rgba[i, :]
    ats = res.atoms
    for at in ats:
        at.color = rgba[i, :]

# run(session, 'hkcage 1 0 alpha hexagonal-dual radius ' + str(radius) + ' spherefactor 0.2')
run(session, 'view orient')
run(session, 'lighting full')
run(session, f'save {pdb}_capsid.png')






def read_config(file_path):
    import toml
    params_dict = toml.load(file_path)

    return params_dict


def create_directories(params_dict):
    from pathlib import Path

    # save_params = ['save_pdb_path', 'save_cg_path', 'save_mode_path', 'save_bfactors_path', 'save_results_path']
    if 'save_pdb_path' not in params_dict['PDB'].keys():
        params_dict['PDB']['save_pdb_path'] = '/'
    if 'save_cg_path' not in params_dict['CG'].keys():
        params_dict['CG']['save_cg_path'] = '/'
    if 'save_mode_path' not in params_dict['NMA'].keys():
        params_dict['NMA']['save_mode_path'] = '/'
    if 'save_bfactors_path' not in params_dict['b_factors'].keys():
        params_dict['b_factors']['save_bfactors_path'] = '/'
    if 'save_results_path' not in params_dict['QRC'].keys():
        params_dict['QRC']['save_results_path'] = '/'

    save_suffix = '_path'
    params = params_dict.copy()

    top_path = params['PDB']['save_all_path']
    Path(top_path).mkdir(parents=True, exist_ok=True)

    # This needs refactoring but it does its job of creating all the folders for saving results

    for k, v in params.items():
        for key, val in v.items():
            if key.endswith(save_suffix) and not key == 'save_all_path' and not key == 'chimerax_path':
                params_dict[k][key] = top_path + val
                Path(params_dict[k][key]).mkdir(parents=True, exist_ok=True)



def run_capsid(params_path):
    from timeit import default_timer as timer
    pycap_start = timer()

    params_dict = read_config(params_path)
    create_directories(params_dict)

    save_all_path = params_dict['PDB'].pop('save_all_path')
    import shutil
    shutil.copy(params_path, save_all_path)

    if 'suppress_plots' in params_dict['plotting'].keys():
        if params_dict['plotting']['suppress_plots']:
            import matplotlib
            matplotlib.use('Agg')


    from .PDB import getCapsid
    pdb = params_dict['PDB']['pdb']
    capsid, calphas, asymmetric_unit, coords, bfactors, chain_starts, title = getCapsid(**params_dict['PDB'])

    from .CG import buildENMPreset
    if params_dict['CG']['preset']=='bbENM':
        params_dict['CG']['chain_starts'] = chain_starts
    kirch, hessian = buildENMPreset(coords,  **params_dict['CG'])

    from .NMA import modeCalc
    evals, evecs = modeCalc(hessian, **params_dict['NMA'])

    from .NMA import fitCompareBfactors
    evals_scaled, evecs_scaled, cc, gamma, n_modes = fitCompareBfactors(evals, evecs, bfactors, pdb, **params_dict['b_factors'])

    from .NMA import calcDistFlucts
    from .QRC import findQuasiRigidClusters

    dist_flucts = calcDistFlucts(evals_scaled, evecs, coords, **params_dict['QRC'])


    if not 'cluster_stop' in params_dict['QRC'].keys():
        print('No cluster range specified, defaulting to: 4-[number of chains]')
        import biotite.structure as struc
        params_dict['QRC']['cluster_start'] = 4
        params_dict['QRC']['cluster_stop'] = struc.get_chain_count(capsid)+1
        params_dict['QRC']['cluster_step'] = 2

    if 'clust_ignore_chains' in params_dict['QRC'].keys():
        from .QRC import filterChains
        clust_mask = filterChains(calphas, clust_ignore_chains=params_dict['QRC']['clust_ignore_chains'])
        params_dict['QRC'].pop('clust_ignore_chains')
    else:
        clust_mask = None

    labels, score, residue_scores = findQuasiRigidClusters(pdb, dist_flucts, clust_mask=clust_mask, **params_dict['QRC'])

    pycap_time = timer() - pycap_start
    print(f'pyCapsid total execution time for {pdb}: {pycap_time}')

    vis_method = params_dict['VIS']['method']

    from .VIS import visualizeResults

    if vis_method == 'chimerax':
        visualizeResults(pdb, capsid, labels, clust_mask=clust_mask, method=vis_method, chimerax_path=params_dict['VIS']['chimerax_path'])
    elif vis_method == 'nglview':
        from pyCapsid.VIS import createCapsidView
        view_clusters = createCapsidView(pdb, capsid)
        literal = """To visualize results in notebook use the following code in another cell:
        from pyCapsid.VIS import createClusterRepresentation
        createClusterRepresentation(pdb, labels, view_clusters)
        view_clusters
        """
        print(literal)
        return pdb, labels, view_clusters
    else:
        print('method must be chimerax or nglview')





def run_capsid_report(params_path):
    from timeit import default_timer as timer
    pycap_start = timer()

    params_dict = read_config(params_path)
    create_directories(params_dict)
    save_all_path = params_dict['PDB'].pop('save_all_path')
    import shutil
    shutil.copy(params_path, save_all_path)

    if 'suppress_plots' in params_dict['plotting'].keys():
        if params_dict['plotting']['suppress_plots']:
            import matplotlib
            matplotlib.use('Agg')

    from pyCapsid.PDB import getCapsid
    pdb = params_dict['PDB']['pdb']
    capsid, calphas, asymmetric_unit, coords, bfactors, chain_starts, title = getCapsid(**params_dict['PDB'])

    from pyCapsid.CG import buildENMPreset
    if params_dict['CG']['preset'] == 'bbENM':
        params_dict['CG']['chain_starts'] = chain_starts

    if params_dict['CG']['preset'] == 'GNM':
        is3d = False
    else:
        is3d = True

    kirch, hessian = buildENMPreset(coords, **params_dict['CG'])

    from pyCapsid.NMA import modeCalc
    evals, evecs = modeCalc(hessian, is3d=is3d, **params_dict['NMA'])

    from pyCapsid.NMA import fitCompareBfactors
    evals_scaled, evecs_scaled, cc, gamma, n_modes = fitCompareBfactors(evals, evecs, bfactors, pdb, is3d=is3d,
                                                                        **params_dict['b_factors'])

    from pyCapsid.NMA import calcDistFlucts
    from pyCapsid.QRC import findQuasiRigidClusters

    dist_flucts = calcDistFlucts(evals_scaled, evecs, coords, is3d=is3d, **params_dict['QRC'])

    if not 'cluster_stop' in params_dict['QRC'].keys():
        print('No cluster range specified, defaulting to: 4-[number of chains]')
        import biotite.structure as struc
        params_dict['QRC']['cluster_start'] = 4
        params_dict['QRC']['cluster_stop'] = struc.get_chain_count(capsid) + 1
        params_dict['QRC']['cluster_step'] = 2

    if 'clust_ignore_chains' in params_dict['QRC'].keys():
        from pyCapsid.QRC import filterChains
        clust_mask = filterChains(calphas, clust_ignore_chains=params_dict['QRC']['clust_ignore_chains'])
        params_dict['QRC'].pop('clust_ignore_chains')
    else:
        clust_mask = None

    labels, score, residue_scores = findQuasiRigidClusters(pdb, dist_flucts, clust_mask=clust_mask,
                                                           **params_dict['QRC'])

    pycap_time = timer() - pycap_start
    print(f'pyCapsid total execution time for {pdb}: {pycap_time}')

    vis_method = params_dict['VIS']['method']

    from pyCapsid.VIS import visualizeResults

    createReport(pdb, save_all_path, n_modes, residue_scores, asymmetric_unit, calphas, params_dict['CG']['preset'], cc,
                 gamma, n_modes, pycap_time)

    if vis_method == 'chimerax':
        report_path = f'{save_all_path}/{pdb}_pyCapsid_report'
        from .VIS import chimeraxReportScriptVis
        c_path = params_dict['VIS'].pop('chimerax_path')
        chimeraxReportScriptVis(report_path, chimerax_path=c_path, **params_dict['VIS'])
        import webbrowser
        import os
        filename = 'file:///' + os.path.abspath(f'{report_path}/pyCapsid_report.html')
        webbrowser.open_new_tab(filename)
    elif vis_method == 'nglview':
        from pyCapsid.VIS import createCapsidView
        view_clusters = createCapsidView(pdb, capsid)
        literal = """To visualize results in notebook use the following code in another cell:
        from pyCapsid.VIS import createClusterRepresentation
        createClusterRepresentation(pdb, labels, view_clusters)
        view_clusters
        """
        print(literal)
        return pdb, labels, view_clusters
    else:
        print('method must be chimerax or nglview')


def createReport(pdb, save_all_path, n_modes_best, residue_scores, asymmetric_unit, calphas, ENM_model, cc, gamma,
                 n_modes, pycap_time):
    # Prepare file names
    file_name = 'pyCapsid_report'
    file_md = file_name + '.md'  # Markdown
    file_html = file_name + '.html'  # HTML
    file_docx = file_name + '.docx'

    # Include libraries to generate plots
    import matplotlib.pyplot as plt
    import seaborn as sns
    import pandas as pd
    import numpy as np

    import os
    parent_dir = save_all_path
    report_dir_name = f'{pdb}_pyCapsid_report'
    report_dir = parent_dir + '/' + report_dir_name

    if not os.path.exists(report_dir):
        os.mkdir(report_dir)

    # Generate figures folder
    figures_dir_name = 'figures'
    figures_dir = report_dir + '/' + figures_dir_name

    if not os.path.exists(figures_dir):
        os.mkdir(figures_dir)

    # Output files (quantiative plots)

    # Correlation coefficients (CC) across modes
    cc_modes_name_out = 'cc_modes'
    cc_modes_fig_png = cc_modes_name_out + '.png'
    cc_modes_fig_svg = cc_modes_name_out + '.svg'
    cc_modes_fig_eps = cc_modes_name_out + '.eps'
    cc_modes_fig_pdf = cc_modes_name_out + '.pdf'
    cc_modes_data_csv = cc_modes_name_out + '.csv'
    cc_modes_dir = figures_dir + '/' + cc_modes_name_out
    if not os.path.exists(cc_modes_dir):
        os.mkdir(cc_modes_dir)

    # b-factors
    b_factors_name_out = 'b_factors'
    b_factors_fig_png = b_factors_name_out + '.png'
    b_factors_fig_svg = b_factors_name_out + '.svg'
    b_factors_fig_eps = b_factors_name_out + '.eps'
    b_factors_fig_pdf = b_factors_name_out + '.pdf'

    b_factors_fit_name_out = 'b_factors_fit'
    b_factors_fit_fig_png = b_factors_fit_name_out + '.png'
    b_factors_fit_fig_svg = b_factors_fit_name_out + '.svg'
    b_factors_fit_fig_eps = b_factors_fit_name_out + '.eps'
    b_factors_fit_fig_pdf = b_factors_fit_name_out + '.pdf'

    b_factors_data_csv = b_factors_name_out + '.csv'
    b_factors_dir = figures_dir + '/' + b_factors_name_out
    if not os.path.exists(b_factors_dir):
        os.mkdir(b_factors_dir)

    # Cluster quality score (CQS)
    cluster_quality_name_out = 'cluster_quality'
    cluster_quality_fig_png = cluster_quality_name_out + '.png'
    cluster_quality_fig_svg = cluster_quality_name_out + '.svg'
    cluster_quality_fig_eps = cluster_quality_name_out + '.eps'
    cluster_quality_fig_pdf = cluster_quality_name_out + '.pdf'
    cluster_quality_data_csv = cluster_quality_name_out + '.csv'
    cluster_quality_dir = figures_dir + '/' + cluster_quality_name_out
    if not os.path.exists(cluster_quality_dir):
        os.mkdir(cluster_quality_dir)

    # Cluster quality score colorbar
    cluster_quality_colorbar_name_out = 'quality_score_colorbar'
    cluster_quality_colorbar_fig_svg = cluster_quality_colorbar_name_out + '.svg'
    cluster_quality_colorbar_fig_png = cluster_quality_colorbar_name_out + '.png'

    # Generate correlation coefficients dataframe and plots

    ## Upload and store file
    dir_loc = parent_dir
    file_name = 'CC_by_mode.npz'
    path_target = dir_loc + '/' + file_name
    cc_modes = np.load(path_target)

    ## Generate and save data frame
    dir_loc = cc_modes_dir
    file_name = cc_modes_data_csv
    path_target = dir_loc + '/' + file_name

    npz = cc_modes
    df = pd.DataFrame(data=[npz['mode_indices']+1, npz['data']]).T
    df.columns = ['Modes', 'CC']
    df_cc_modes = df
    df_cc_modes.to_csv(path_target, index=False)
    cc_best = df_cc_modes[df_cc_modes['Modes'] == n_modes_best+1]['CC'].values[0]

    ## Generate plots
    plt.ioff()  # Comment this command to display figure in the notebook
    sns.set_style('darkgrid')
    fig, ax = plt.subplots(1, 1)
    sns.lineplot(ax=ax, x='Modes', y='CC', data=df_cc_modes)
    ax.set_title('Agreement between predicted and empirical B-factors')
    ax.set_xlabel('Number of modes')
    ax.set_ylabel('Correlation coefficient (CC)')
    ax.vlines(x=n_modes_best+1, ymin=0, ymax=cc_best, color='black')
    # plt.show(fig) ## Comment out to display figure in notebook

    ## Save plots
    dir_loc = cc_modes_dir
    ### PNG
    file_name = cc_modes_fig_png
    path_target = dir_loc + '/' + file_name
    fig.savefig(path_target, format='png', dpi=300)
    ### EPS
    file_name = cc_modes_fig_eps
    path_target = dir_loc + '/' + file_name
    fig.savefig(path_target, format='eps', bbox_inches='tight')
    ### SVG
    file_name = cc_modes_fig_svg
    path_target = dir_loc + '/' + file_name
    fig.savefig(path_target, format='svg')
    ### PDF
    file_name = cc_modes_fig_pdf
    path_target = dir_loc + '/' + file_name
    fig.savefig(path_target, format='pdf')

    ## Generate caption
    text = '**Figure: Selection of the best number of modes** (below). Correlation coefficient obtained comparing the experimental B-factors with respect the B-factors predicted an increasing number of low frequency modes. The black line higlights the number of modes that yield the optimal correlation coefficient with the empirical B-factors.'
    caption_cc_modes = text

    # Generate b-factors dataframe and plots

    ## Upload and store file
    dir_loc = parent_dir
    file_name = 'b_factors.npz'
    path_target = dir_loc + '/' + file_name
    b_factors = np.load(path_target)

    ## Generate and save data frame
    ## (!! revise method for residues in asymmetric unit)
    dir_loc = b_factors_dir
    file_name = b_factors_data_csv
    path_target = dir_loc + '/' + file_name

    npz = b_factors
    k = npz['k']
    unscaled = npz['bfactors_predicted'] / k
    x_range = np.arange(unscaled.min(), unscaled.max(), 0.01)
    y_line = x_range * k
    df_bfactors = pd.DataFrame(
        data=[npz['residue_numbers'], npz['bfactors_experimental'], npz['bfactors_predicted'], unscaled, x_range,
              y_line]).T
    df_bfactors.columns = ['Residue', 'Experimental', 'Predicted', 'Unscaled', 'x_line', 'y_line']
    n_res_asym = int(len(df_bfactors) / 60)  # ! residues in asymmetric unit (refactor)
    df_bfactors_asym = df_bfactors[0:n_res_asym]
    # len(df_bfactors_asym)
    df_bfactors_asym.to_csv(path_target, index=False)

    ## Generate plots
    df = df_bfactors_asym
    plt.ioff()  # Comment this command to display figure in the notebook
    sns.set_style('darkgrid')
    fig, ax = plt.subplots(1, 1, figsize=(6, 4), sharex=False)

    # b-factor profile
    sns.lineplot(ax=ax, x='Residue', y='Predicted', data=df, label='Predicted')
    sns.lineplot(ax=ax, x='Residue', y='Experimental', data=df, label='Experimental')
    ax.set_title('Protein complex\'s vibration profile (asymmetric unit)')
    ax.set_xlabel('Residue number')
    ax.set_ylabel(r"B-factor ($\AA^2$)")
    ax.legend(frameon=False)

    # plt.show(fig) ## Comment out to display figure in notebook

    ## Save plots
    dir_loc = b_factors_dir
    ### PNG
    file_name = b_factors_fig_png
    path_target = dir_loc + '/' + file_name
    fig.savefig(path_target, format='png', dpi=300)
    ### EPS
    file_name = b_factors_fig_eps
    path_target = dir_loc + '/' + file_name
    fig.savefig(path_target, format='eps', bbox_inches='tight')
    ### SVG
    file_name = b_factors_fig_svg
    path_target = dir_loc + '/' + file_name
    fig.savefig(path_target, format='svg')
    ### PDF
    file_name = b_factors_fig_pdf
    path_target = dir_loc + '/' + file_name
    fig.savefig(path_target, format='pdf')

    ## Generate caption
    text = '**Figure: Correlation of B-factors** (below). Empirical (blue) and predicted (orange) B-factors for each residue in the asymmetric unit.'
    caption_bfactors = text

    # Generate b-factors dataframe and plots

    ## Upload and store file
    dir_loc = parent_dir
    file_name = 'b_factors.npz'
    path_target = dir_loc + '/' + file_name
    b_factors = np.load(path_target)

    ## Generate and save data frame
    ## (!! revise method for residues in asymmetric unit)
    dir_loc = b_factors_dir
    file_name = b_factors_data_csv
    path_target = dir_loc + '/' + file_name

    npz = b_factors
    k = npz['k']
    unscaled = npz['bfactors_predicted'] / k
    x_range = np.arange(unscaled.min(), unscaled.max(), 0.01)
    y_line = x_range * k
    df_bfactors = pd.DataFrame(
        data=[npz['residue_numbers'], npz['bfactors_experimental'], npz['bfactors_predicted'], unscaled, x_range,
              y_line]).T
    df_bfactors.columns = ['Residue', 'Experimental', 'Predicted', 'Unscaled', 'x_line', 'y_line']
    n_res_asym = int(len(df_bfactors) / 60)  # ! residues in asymmetric unit (refactor)
    df_bfactors_asym = df_bfactors[0:n_res_asym]
    # len(df_bfactors_asym)
    df_bfactors_asym.to_csv(path_target, index=False)

    ## Generate plots
    df = df_bfactors
    plt.ioff()  # Comment this command to display figure in the notebook
    sns.set_style('darkgrid')
    fig, ax = plt.subplots(1, 1, figsize=(6, 4.5))

    # b-factor fit
    sns.scatterplot(ax=ax, x='Unscaled', y='Experimental', data=df, label='Scatter')
    sns.lineplot(ax=ax, x='x_line', y='y_line', data=df, label='Linear Fit')
    ax.set_title('ENM Calibration (linear fit)')
    ax.set_xlabel('Predicted squared fluctuations for $\gamma  = 1$ ($\AA^2$)')
    ax.set_ylabel(r"B-factor ($\AA^2$)")
    ax.legend(frameon=False)

    # plt.show(fig) ## Comment out to display figure in notebook

    ## Save plots
    dir_loc = b_factors_dir
    ### PNG
    file_name = b_factors_fit_fig_png
    path_target = dir_loc + '/' + file_name
    fig.savefig(path_target, format='png', dpi=300)
    ### EPS
    file_name = b_factors_fit_fig_eps
    path_target = dir_loc + '/' + file_name
    fig.savefig(path_target, format='eps', bbox_inches='tight')
    ### SVG
    file_name = b_factors_fit_fig_svg
    path_target = dir_loc + '/' + file_name
    fig.savefig(path_target, format='svg')
    ### PDF
    file_name = b_factors_fit_fig_pdf
    path_target = dir_loc + '/' + file_name
    fig.savefig(path_target, format='pdf')

    ## Generate caption
    text = '**Figure: Model calibration** (below). The empirical B-factors are plotted against the predicted fluctuations. A regression line was fitted. From the slope (a), the calibration constant (gamma) was obtained using the formula: *gamma = (8(pi)^2)/(3a)*.'
    caption_bfactors_fit = text

    # Generate quality score dataframe and plots

    ## Obtain optimal number of clusters
    dir_loc = parent_dir
    file_name = str(pdb) + '_final_results.npz'
    path_target = dir_loc + '/' + file_name
    npz_loaded = np.load(path_target)
    nc = int(npz_loaded['nc'])

    ## Upload and store quality score file
    dir_loc = parent_dir
    file_name = str(pdb) + '_final_results_full.npz'
    path_target = dir_loc + '/' + file_name
    npz_loaded = np.load(path_target)

    ## Generate and save data frame
    dir_loc = cluster_quality_dir
    file_name = cluster_quality_data_csv
    path_target = dir_loc + '/' + file_name

    df = pd.DataFrame(data=[npz_loaded['nc_range'], npz_loaded['score'], npz_loaded['numtypes']]).T
    df.columns = ['Clusters', 'Quality score', 'Unique']
    df.to_csv(path_target, index=False)

    # Obtain score and unique clustering for optimal number of clusters (nc)
    ii = df[df['Clusters'] == nc].index[0]
    score = df['Quality score'].iloc[ii]
    unique = df['Unique'].iloc[ii]

    ## Generate and save plots
    plt.ioff()  # Comment this command to display figure in the notebook
    sns.set_style('darkgrid')
    fig, ax = plt.subplots(2, 1, figsize=(6, 3), sharex=True)
    fig.suptitle('Cluster quality score analysis')
    sns.lineplot(ax=ax[0], x='Clusters', y='Quality score', data=df)
    sns.lineplot(ax=ax[1], x='Clusters', y='Unique', data=df)
    ax[1].set_xlabel('Number of clusters')
    ax[0].set_ylabel('Quality score')
    ax[1].set_ylabel('Unique clusters')
    ax[0].vlines(x=nc, ymin=0, ymax=score, color='black')
    ymax = ax[1].get_ylim()[1]
    ax[1].vlines(x=nc, ymin=unique, ymax=ymax, color='black')
    fig.tight_layout()
    # plt.show(fig) ## Comment out to display figure in notebook

    ## Save plots
    dir_loc = cluster_quality_dir
    ### PNG
    file_name = cluster_quality_fig_png
    path_target = dir_loc + '/' + file_name
    fig.savefig(path_target, format='png', dpi=300)
    ### EPS
    file_name = cluster_quality_fig_eps
    path_target = dir_loc + '/' + file_name
    fig.savefig(path_target, format='eps', bbox_inches='tight')
    ### SVG
    file_name = cluster_quality_fig_svg
    path_target = dir_loc + '/' + file_name
    fig.savefig(path_target, format='svg')
    ### PDF
    file_name = cluster_quality_fig_pdf
    path_target = dir_loc + '/' + file_name
    fig.savefig(path_target, format='pdf')

    ## Generate caption
    text = '**Figure: Optimal clustering selection** (below). Quality score (top) and number of unique clusters (bottom) obtained as a function of the number of clusters tested. The black lines in each plot higlight the quality score and number of unique clusters for the optimal number of clusters identified.'
    caption_cluster_quality = text

    # Quality score colorbar figure
    import matplotlib.colorbar as colorbar
    import matplotlib.pyplot as plt
    from pyCapsid.VIS import clusters_colormap_hexcolor
    import numpy as np
    hexcolor, cmap = clusters_colormap_hexcolor(residue_scores, rwb_scale=True)
    fig, ax = plt.subplots(figsize=(10, 1))
    cb = colorbar.ColorbarBase(ax, orientation='horizontal',
                               cmap=cmap, norm=plt.Normalize(np.min(residue_scores), np.max(residue_scores)))
    # plt.show()
    fig.suptitle('Clustering quality scores ranging from low (red) to high (blue)', y=-0.02)
    fig.tight_layout()
    fig.savefig(cluster_quality_colorbar_fig_svg)
    fig.savefig(f'{cluster_quality_dir}/{cluster_quality_colorbar_fig_svg}')

    # Structures folder
    # Generate folder
    structures_dir = figures_dir + '/' + 'structures'
    if not os.path.exists(structures_dir):
        os.mkdir(structures_dir)

    # Set structure view
    asym_unit_name_out = f'{pdb}_asymmetric_unit'
    asym_unit_fig_png = asym_unit_name_out + '.png'

    # Save image
    dir_loc = structures_dir
    file_name = asym_unit_fig_png
    path_target = dir_loc + '/' + file_name
    from PIL import Image
    image = Image.new('RGB', (20, 20))
    image.save(path_target)

    # Set structure view
    capsid_name_out = f'{pdb}_full_capsid'
    capsid_fig_png = capsid_name_out + '.png'

    # Save image
    dir_loc = structures_dir
    file_name = capsid_fig_png
    path_target = dir_loc + '/' + file_name

    from PIL import Image
    image = Image.new('RGB', (20, 20))
    image.save(path_target)

    # Set structure view
    capsid_scores_name_out = f'{pdb}_residue_cluster_scores'
    capsid_scores_fig_png = capsid_scores_name_out + '.png'

    # Save image
    dir_loc = structures_dir
    file_name = capsid_scores_fig_png
    path_target = dir_loc + '/' + file_name

    from PIL import Image
    image = Image.new('RGB', (20, 20))
    image.save(path_target)

    capsid_clusters_name_out = f'{pdb}_highest_quality_clusters'
    capsid_clusters_fig_png = capsid_clusters_name_out + '.png'

    # Save image
    dir_loc = structures_dir
    file_name = capsid_clusters_fig_png
    path_target = dir_loc + '/' + file_name

    from PIL import Image
    image = Image.new('RGB', (20, 20))
    image.save(path_target)

    chimerax_dir = report_dir + "/chimerax/"
    if not os.path.exists(chimerax_dir):
        os.mkdir(chimerax_dir)

    import pyCapsid.scripts as scpath
    import os
    import shutil

    script_path = os.path.abspath(scpath.__file__)
    script_path_1 = script_path.replace('__init__', 'chimerax_script_colab')
    script_path_2 = script_path.replace('__init__', 'chimerax_script_animate_mode')
    shutil.copy(script_path_1, chimerax_dir)
    shutil.copy(script_path_2, chimerax_dir)
    shutil.copy(f'{save_all_path}/{pdb}_final_results_full.npz', chimerax_dir)
    shutil.copy(f'{save_all_path}/{pdb}_final_results.npz', chimerax_dir)
    # shutil.copy('modes.npz', chimerax_dir + f'{pdb}_modes.npz')
    # if pdb_source == 'upload':
    #   shutil.copy(pdb_file_name, chimerax_dir)

    # Include only the first non-degenerate eigenmode in the chimeras folder
    modes_saved = np.load(f'{save_all_path}/modes.npz')
    evals_saved = modes_saved['eigen_vals']
    evecs_saved = modes_saved['eigen_vecs']
    uniques, inds, counts = np.unique(evals_saved.round(decimals=8), return_index=True, return_counts=True)
    icoEvalInds = inds[counts == 1]
    print('Indices of non-degenerate eigenmodes: ', icoEvalInds)

    if len(icoEvalInds) == 0:
        print('No non-degenerate modes found, including lowest-frequency mode for visualization')
        n_mode = 0
    else:
        print('Including modes up to lowest frequency non-degenerate mode for visualization')
        n_mode = icoEvalInds[0] + 1

    np.savez_compressed(chimerax_dir + f'{pdb}_modes.npz', eigen_vals=evals_saved[:n_mode],
                        eigen_vecs=evecs_saved[:, :n_mode])

    # Write readme
    with open(chimerax_dir + "readme_chimerax.md", 'w') as file:
        line = "# ChimeraX visualization script"
        file.write('\n' + line)
        line = 'To visualize the clustering results in ChimeraX we provide a python script in the pyCapsid report folder. To use this script, after extracting the pyCapsid_report folder, open ChimeraX and run the following command:'
        file.write('\n' + line)
        line = f'``` \n runscript "path/to/pyCapsid_report/chimerax/chimerax_script_colab.py" \n ```'
        file.write('\n' + line)
        line = f'``` Alternatively you can use the following command, which will open a window and  prompt you to find and select the script:'
        file.write('\n' + line)
        line = f'``` \n runscript browse \n ```'
        file.write('\n' + line)
        line = 'This will create the 4 images in the pyCapsid_report/figures/structures/ folder which will be displayed in the report.'
        file.write('\n' + line)
        line = 'The user may also visualize motion of the structure along a given normal mode using a separate script in a similar manner:'
        file.write('\n' + line)
        line = f'``` \n runscript "path/to/pyCapsid_report/chimerax/chimerax_script_animate_mode.py" \n ```'
        file.write('\n' + line)
        line = f'``` \n runscript browse \n ```'
        file.write('\n' + line)
        line = f'This will create an .mp4 file in the figures/structures folder of the report.'
        file.write('\n' + line)
        line = f'Only one mode may be visualized at a time. By default the lowest frequency non-degenerate mode is visualized, and only modes of that frequency and lower are included in the report. To visualize higher frequency modes, you may copy the `modes.npz` file from `pyCapsid_results` to the `chimerax` folder of the report and rename it accordingly.'
        file.write('\n' + line)
        line = f'Both scripts can be modified to fit the users needs. To do this reference the ChimeraX commands (https://www.cgl.ucsf.edu/chimerax/docs/user/index.html) and the ChimeraX Python API (https://www.cgl.ucsf.edu/chimerax/docs/devel/index.html#).'
        file.write('\n' + line)

    # KEY RESULTS DICTIONARY WHICH WILL BE SAVED AS CSV
    key_results = {}

    # Generate markdown output file
    dir_loc = report_dir
    file_name = file_md
    path_target = dir_loc + '/' + file_name
    f = open(path_target, 'w')

    # Load data needed
    from datetime import date
    today = date.today()

    # HEADING
    text = '# pyCapsid Report'
    f.write('\n' + text)

    text = today.strftime("%B %d, %Y")
    f.write('\n' + text)

    # INPUT STRUCTURE
    f.write('\n')
    text = '## Input structure'
    f.write('\n' + text)

    text = 'Identifier: ' + str(pdb)
    f.write('\n' + text)

    f.write('\n')
    import biotite.structure as struc
    n_asym_residues = struc.get_residue_count(asymmetric_unit)
    key_results['n_asym_residues'] = n_asym_residues
    text = 'Number of residues in the asymmetric unit: ' + str(n_asym_residues)
    f.write('\n' + text)

    f.write('\n')
    n_asym_chains = struc.get_chain_count(asymmetric_unit)
    key_results['n_asym_chains'] = n_asym_chains
    text = 'Number of protein chains in the asymmetric unit: ' + str(n_asym_chains)
    f.write('\n' + text)

    f.write('\n')
    n_full_residues = len(calphas)
    asym_factor = int(n_full_residues / n_asym_residues)
    text = 'Multiplying factor to generate the complete protein complex: ' + str(asym_factor)
    f.write('\n' + text)

    f.write('\n')
    text = '+ If the multiplying factor is one (*m = 1*), the protein complex and the asymmetric unit are the same.'
    f.write('\n' + text)

    f.write('\n')
    text = '+ If the multiplying factor is larger than one (*m > 1*), the protein complex is *m* times the asymmetric unit.'
    f.write('\n' + text)

    f.write('\n')
    key_results['n_full_residues'] = n_full_residues
    text = 'Number of residues in the protein complex: ' + str(n_full_residues)
    f.write('\n' + text)

    f.write('\n')
    n_full_chains = int(asym_factor * n_asym_chains)
    key_results['n_full_chains'] = n_full_chains
    text = 'Number of protein chains in the protein complex: ' + str(n_full_chains)
    f.write('\n' + text)

    ## Asymmetric unit
    f.write('\n')
    text = '***'
    f.write('\n' + text)
    f.write('\n')
    text = '**Figure: Asymmetric unit** (below). Ribbon diagram of the protein complex\'s asymmetric unit.'
    f.write('\n' + text)
    f.write('\n')
    text = '![Asymmetric unit](' + './figures/structures/' + asym_unit_fig_png + ')'
    f.write('\n' + text)

    ## Full structure
    f.write('\n')
    text = '***'
    f.write('\n' + text)
    f.write('\n')
    text = '**Figure: Full protein complex** (below). Ribbon diagram of the full protein complex.'
    f.write('\n' + text)
    f.write('\n')
    text = '![Full protein complex](' + './figures/structures/' + capsid_fig_png + ')'
    f.write('\n' + text)
    f.write('\n')
    text = '***'
    f.write('\n' + text)

    # Elastic network model
    f.write('\n')
    text = '## Elastic network model'
    f.write('\n' + text)

    f.write('\n')
    key_results['ENM_model'] = ENM_model
    text = 'Elastic model used: ' + str(ENM_model)
    f.write('\n' + text)

    f.write('\n')
    key_results['gamma'] = str(round(gamma, 2))
    text = 'Calibrated stiffness constant (gamma): ' + str(round(gamma, 2))
    f.write('\n' + text)
    f.write('\n')
    text = '+ This constant was fitted to scale the model to the structure, assuming a linear relationship between the residues fluctuations and B-factors.'
    f.write('\n' + text)

    f.write('\n')
    text = '***'
    f.write('\n' + text)
    f.write('\n')
    text = caption_bfactors_fit
    f.write('\n' + text)
    f.write('\n')
    ### Path to render figure by user locally in markdown
    subdir_sep = b_factors_dir.split(sep='/')[::-1][0:2][::-1]
    subdir = './' + subdir_sep[0] + '/' + subdir_sep[1]
    text = '![Model calibration](' + subdir + '/' + b_factors_fit_fig_svg + ')'
    f.write('\n' + text)
    ### Path to render figure by pypandoc in Colab
    dir_loc = b_factors_dir
    file_name = b_factors_fit_fig_svg
    path_target = dir_loc + '/' + file_name
    text = '![](' + path_target + ')'
    f.write('\n' + text)

    # Normal modes and B-factors
    CC = float(b_factors['CC'])

    f.write('\n')
    text = '## Normal mode analysis (NMA)'
    f.write('\n' + text)

    f.write('\n')
    key_results['n_modes_optimal'] = n_modes_best + 1
    text = 'Optimal number of modes reproducing B-factors: ' + str(n_modes_best + 1)
    f.write('\n' + text)

    f.write('\n')
    key_results['CC'] = CC
    text = 'Correlation between empirical and predicted B-factors: ' + str(round(CC, 2))
    f.write('\n' + text)

    ## CC vs number of modes figure
    f.write('\n')
    text = '***'
    f.write('\n' + text)
    f.write('\n')
    text = caption_cc_modes
    f.write('\n' + text)
    f.write('\n')
    ### Path to render figure by user locally in markdown
    subdir_sep = cc_modes_dir.split(sep='/')[::-1][0:2][::-1]
    subdir = './' + subdir_sep[0] + '/' + subdir_sep[1]
    text = '![Selection of the best number of modes](' + subdir + '/' + cc_modes_fig_svg + ')'
    f.write('\n' + text)
    ### Path to render figure by pypandoc in Colab
    dir_loc = cc_modes_dir
    file_name = cc_modes_fig_svg
    path_target = dir_loc + '/' + file_name
    text = '![](' + path_target + ')'
    f.write('\n' + text)

    ## B-factors figure
    f.write('\n')
    text = caption_bfactors
    f.write('\n' + text)

    f.write('\n')
    ### Path to render figure by user locally in markdown
    subdir_sep = b_factors_dir.split(sep='/')[::-1][0:2][::-1]
    subdir = './' + subdir_sep[0] + '/' + subdir_sep[1]
    text = '![Correlation of B-factors](' + subdir + '/' + b_factors_fig_svg + ')'
    f.write('\n' + text)
    ### Path to render figure by pypandoc in Colab
    dir_loc = b_factors_dir
    file_name = b_factors_fig_svg
    path_target = dir_loc + '/' + file_name
    text = '![](' + path_target + ')'
    f.write('\n' + text)

    f.write('\n')
    text = '***'
    f.write('\n' + text)

    # Quasi-rigid units
    ## Upload and store file
    dir_loc = parent_dir
    file_name = str(pdb) + '_final_results.npz'
    path_target = dir_loc + '/' + file_name
    npz_loaded = np.load(path_target)
    nc = int(npz_loaded['nc'])

    f.write('\n')
    text = '## Quasi-rigid mechanical units'
    f.write('\n' + text)

    f.write('\n')
    key_results['n_clusters_optimal'] = nc
    text = 'Number of optimal quasi-rigid mechanical units identified: ' + str(nc)
    f.write('\n' + text)

    # f.write('\n')
    # text = '!!! Add figure plotting the quality score and number of unique clusters as a function of the number of selected clusters. Include captions following standard publication style.'
    # f.write('\n'+text)

    ## Cluster quality figure
    f.write('\n')

    f.write('\n')
    text = caption_cluster_quality
    f.write('\n' + text)

    ### Path to render figure by user locally in markdown
    subdir_sep = cluster_quality_dir.split(sep='/')[::-1][0:2][::-1]
    subdir = './' + subdir_sep[0] + '/' + subdir_sep[1]
    text = '![](' + subdir + '/' + cluster_quality_fig_svg + ')'
    f.write('\n' + text)
    ### Path to render figure by pypandoc in Colab
    dir_loc = cluster_quality_dir
    file_name = cluster_quality_fig_svg
    path_target = dir_loc + '/' + file_name
    text = '![](' + path_target + ')'
    f.write('\n' + text)

    f.write('\n')
    text = '***'
    f.write('\n' + text)
    f.write('\n')
    text = '**Figure: Quasi-rigid clusters** (below). Ribbon representation of the complete structure, with each residue colored according to its cluster membership. Residues with the same color are members of the same quasi-rigid cluster.'
    f.write('\n' + text)
    f.write('\n')
    text = '![](' + './figures/structures/' + capsid_clusters_fig_png + ')'
    f.write('\n' + text)

    # f.write('\n')
    # text = '!!! Add figure displaying a representative of each unique cluster. Include caption following standard publication style.'
    # f.write('\n'+text)

    f.write('\n')
    text = '**Figure: Residue cluster quality score** (below). Ribbon representation of the complete structure, with each residue colored according to its cluster quality score.  This is a measure of how rigid each residue is with respect to its cluster. Blue residues make up the cores of rigid clusters, and red residues represent borders between clusters.'
    f.write('\n' + text)

    ## clusters
    f.write('\n')
    text = '![](' + './figures/structures/' + capsid_scores_fig_png + ')'
    f.write('\n' + text)

    # colorbar
    f.write('\n')
    text = '![](' + './figures/cluster_quality/' + cluster_quality_colorbar_fig_svg + ')'
    f.write('\n' + text)

    # f.write('\n')
    # text = '!!! Add figure displaying the quality score of each residue in the representative unique clusters. Include caption following standard publication style.'
    # f.write('\n'+text)

    f.write('\n')
    text = "# Missing images in the report? \n "
    f.write('\n' + text)

    f.write('\n')
    text = "To render the missing images locally using ChimeraX, check the 'chimerax' folder of the report, and follow the instructions provided in the readme file. This will automatically place the images in the .md and .html versions. For the word document, copy and paste the relevant images directly into the document."
    f.write('\n' + text)

    f.write('\n')
    text = "If you were visualizing the results using NGLView, you should have several images downloaded corresponding to the relevant images. After extracting \"pyCapsid_report.zip\", place the corresponding images from your download folder into \"pyCapsid_report/figures/structures/\" replacing the empty images in that folder. This will fix the issue for the .md and .html versions. For the word document, copy and paste the relevant images directly into the document."
    f.write('\n' + text)

    # Close ouptut file
    f.close()

    # WRITE KEY_RESULTS CSV
    key_results['main_pycap_time'] = pycap_time
    import csv
    with open(f'{report_dir}/pyCapsid_key_results.csv', 'w') as csv_file:
        writer = csv.writer(csv_file)
        for key, value in key_results.items():
            writer.writerow([key, value])

    import markdown
    # Generate HTML version
    dir_loc = report_dir
    file_name = file_md
    path_in = dir_loc + '/' + file_name
    file_name = file_html
    path_out = dir_loc + '/' + file_name

    markdown.markdownFromFile(input=path_in, output=path_out)


"""Install script for setuptools."""

from setuptools import find_packages
from setuptools import setup
long_description = open("README.md").read()
setup(
    name='pyCapsid',
    version='0.6.0',
    description='A set of computational tools written in python for the analysis of viral capsids',
    long_description =long_description,
    long_description_content_type='text/markdown',
    author='Luque Lab, Colin Brown, Anuradha Agarwal,Jose Bernard Sedalo',
    author_email='colintravisbrown@gmail.com',
    license='MIT License',
    url='https://github.com/luquelab/pycapsid',
    packages=find_packages("src"),
    package_dir={"": "src"},
    python_requires = '>=3.7, <3.13',
    install_requires=[
        'biotite',
        'scikit-learn',
        'numpy',
        'scipy',
        'matplotlib',
        'pillow',
        'numba>=0.57',
        'statsmodels',
        'toml',
        'markdown',
        'seaborn'
    ],
    classifiers=[
        'Development Status :: 1 - Planning',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        "Operating System :: Microsoft :: Windows",
        "Operating System :: MacOS",
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
        'Topic :: Scientific/Engineering :: Bio-Informatics',
    ],
)



"""Functions for visualizing results in chimeraX

"""


def visualizeResults(pdb, capsid, labels, method='chimerax', **kwargs):
    if method == 'chimerax':
        chimeraxViz(labels, pdb, **kwargs)
    elif method == 'nglview':
        return view_pdb_ngl(pdb, capsid, labels, **kwargs)
    else:
        print('method must be chimerax or nglview')

def chimeraxViz(labels, pdb, remote=True, chimerax_path=None, pdb_path='.', save_path='.',
                rwb_scale=False, clust_mask=None, **kwargs):
    """Launches ChimeraX and runs a script that visualizes the results.

    :param labels:
    :param pdb:
    :param remote:
    :param chimerax_path:
    :param pdb_path:
    :param save_path:
    :param script_path:
    :param labels_path:
    :return:
    """

    import os
    import platform
    if chimerax_path is None:
        print('No chimerax path specified, checking in default locations for your OS')
        if platform.system()=='Linux':
            chimerax_path = 'usr/bin/chimerax'
        elif platform.system()=='Windows':
            chimerax_path = 'C:\\Program Files\\ChimeraX\\bin\\ChimeraX.exe'
        elif platform.system()=='Darwin':
            chimerax_path = '/Applications/ChimeraX'
        else:
            print('No chimerax path is given and cannot check default locations')
            chimerax_path = ''

    import matplotlib as mpl
    norm = mpl.colors.Normalize(vmin=np.min(labels), vmax=np.max(labels))

    if rwb_scale:
        import matplotlib.pyplot as plt
        cmap = plt.get_cmap('coolwarm_r')
        rgba = cmap(norm(labels)) * 255
    else:
        cmap = generate_colormap(int(np.max(labels)))
        rgba = cmap(norm(labels)) * 255

    if clust_mask is not None:
        rgba = addUnclusteredColor(rgba, clust_mask)

    from numpy import save
    from tempfile import NamedTemporaryFile
    with NamedTemporaryFile(suffix='.npy', delete=False) as temp_file:
        save(temp_file, rgba)
        rgba_path = os.path.normpath(temp_file.name)
        rgba_path = rgba_path.replace('\\', '/') #os.path.abspath(temp_file.name)
        print(rgba_path)

    # get path to chimerax script
    import pyCapsid.scripts as scpath
    import os

    script_path = os.path.abspath(scpath.__file__)
    script_path = script_path.replace('__init__', 'chimerax_script')

    chimerax_exe = chimerax_path #+ '\\ChimeraX.exe'
    if platform.system()=='Windows':
        cmd_string = f'""{chimerax_exe}" --script "{script_path} {rgba_path} {save_path} {pdb_path} {str(remote)} {pdb} {str(rwb_scale)}""'
    elif platform.system()=='Linux':
        cmd_string = f'{chimerax_exe} --script "{script_path} {rgba_path} {save_path} {pdb_path} {str(remote)} {pdb} {str(rwb_scale)}"'
    else:
        cmd_string = f'{chimerax_exe} --script "{script_path} {rgba_path} {save_path} {pdb_path} {str(remote)} {pdb} {str(rwb_scale)}"'
    print(cmd_string)
    os.system(cmd_string)
    temp_file.close()
    os.unlink(temp_file.name)

def visualizeSavedResults(pdb, results_file, n_cluster=None, method='chimerax', **kwargs):
    import numpy as np
    results = np.load(results_file)
    labels = results['labels']

    nc_range = results['nc_range']
    scores = results['score']

    if n_cluster is None:
        print('Defaulting to highest score clustering')
        ind = np.argmax(scores)
        n_c = nc_range[ind]
    else:
        ind = np.argwhere(nc_range == n_cluster)[0][0]
        n_c = n_cluster

    print(f'Visualizing cluster results of {pdb} for {n_c} clusters')
    labels = labels[ind]
    print(labels)
    if method=='chimerax':
        chimeraxViz(labels, pdb, **kwargs)
    elif method=='nglview':
        view = createCapsidView(pdb)
        return labels, view
    else:
        print('Method must be one of chimerax or nglview')


def addUnclusteredColor(rgba, clust_mask):
    from numpy import insert
    print(rgba.shape)
    print(rgba)
    black = np.array([0,0,0,255])
    for i,val in enumerate(clust_mask):
        if not val:
            rgba = insert(rgba, i, black, axis=0)
    print(rgba.shape)
    print(rgba)
    return rgba



# Adapted from py3dmol tutorial
# https://william-dawson.github.io/using-py3dmol.html
class Atom(dict):
    def __init__(self, line):
        self["type"] = line[0:6].strip()
        self["idx"] = line[6:11].strip()
        self["name"] = line[12:16].strip()
        self["resname"] = line[17:20].strip()
        self["resid"] = int(int(line[22:26]))
        self["x"] = float(line[30:38])
        self["y"] = float(line[38:46])
        self["z"] = float(line[46:54])
        self["sym"] = line[76:78].strip()

    def __str__(self):
        line = list(" " * 80)

        line[0:6] = self["type"].ljust(6)
        line[6:11] = self["idx"].ljust(5)
        line[12:16] = self["name"].ljust(4)
        line[17:20] = self["resname"].ljust(3)
        line[22:26] = str(self["resid"]).ljust(4)
        line[30:38] = str(self["x"]).rjust(8)
        line[38:46] = str(self["y"]).rjust(8)
        line[46:54] = str(self["z"]).rjust(8)
        line[76:78] = self["sym"].rjust(2)
        return "".join(line) + "\n"


class Molecule(list):
    def __init__(self, file):
        for line in file:
            if "ATOM" in line or "HETATM" in line:
                self.append(Atom(line))

    def __str__(self):
        outstr = ""
        for at in self:
            outstr += str(at)

        return outstr
    
# Colormap generation taken from stackexchange
# https://stackoverflow.com/questions/42697933/colormap-with-maximum-distinguishable-colours
def generate_colormap(number_of_distinct_colors: int = 80):
    import math

    import numpy as np
    from matplotlib.colors import ListedColormap
    from matplotlib.cm import hsv

    if number_of_distinct_colors == 0:
        number_of_distinct_colors = 80

    number_of_shades = 7
    if number_of_distinct_colors < (number_of_shades-1):
        number_of_shades = number_of_distinct_colors - 2

    number_of_distinct_colors_with_multiply_of_shades = int(math.ceil(number_of_distinct_colors / number_of_shades) * number_of_shades)

    # Create an array with uniformly drawn floats taken from <0, 1) partition
    linearly_distributed_nums = np.arange(number_of_distinct_colors_with_multiply_of_shades) / number_of_distinct_colors_with_multiply_of_shades

    # We are going to reorganise monotonically growing numbers in such way that there will be single array with saw-like pattern
    #     but each saw tooth is slightly higher than the one before
    # First divide linearly_distributed_nums into number_of_shades sub-arrays containing linearly distributed numbers
    arr_by_shade_rows = linearly_distributed_nums.reshape(number_of_shades, number_of_distinct_colors_with_multiply_of_shades // number_of_shades)

    # Transpose the above matrix (columns become rows) - as a result each row contains saw tooth with values slightly higher than row above
    arr_by_shade_columns = arr_by_shade_rows.T
    
     # Keep number of saw teeth for later
    number_of_partitions = arr_by_shade_columns.shape[0]

    # Flatten the above matrix - join each row into single array
    nums_distributed_like_rising_saw = arr_by_shade_columns.reshape(-1)

    # HSV colour map is cyclic (https://matplotlib.org/tutorials/colors/colormaps.html#cyclic), we'll use this property
    initial_cm = hsv(nums_distributed_like_rising_saw)

    lower_partitions_half = number_of_partitions // 2
    upper_partitions_half = number_of_partitions - lower_partitions_half

    # Modify lower half in such way that colours towards beginning of partition are darker
    # First colours are affected more, colours closer to the middle are affected less
    lower_half = lower_partitions_half * number_of_shades
    for i in range(3):
        initial_cm[0:lower_half, i] *= np.arange(0.2, 1, 0.8/lower_half)

    # Modify second half in such way that colours towards end of partition are less intense and brighter
    # Colours closer to the middle are affected less, colours closer to the end are affected more
    for i in range(3):
        for j in range(upper_partitions_half):
            modifier = np.ones(number_of_shades) - initial_cm[lower_half + j * number_of_shades: lower_half + (j + 1) * number_of_shades, i]
            modifier = j * modifier / upper_partitions_half
            initial_cm[lower_half + j * number_of_shades: lower_half + (j + 1) * number_of_shades, i] += modifier

    return ListedColormap(initial_cm)
  

import numpy as np
def open_pdb(pdb):
    with open(pdb + "_capsid.pdb") as ifile:
        mol = Molecule(ifile)
    return mol

  
def clusters_colormap_hexcolor(clusters, rwb_scale):
    import matplotlib as mpl
    norm = mpl.colors.Normalize(vmin=np.min(clusters), vmax=np.max(clusters))

    if rwb_scale:
        cmap = mpl.pyplot.get_cmap('coolwarm_r')
        rgba = cmap(norm(clusters))
    else:
        cmap = generate_colormap(int(np.max(clusters)))
        rgba = cmap(norm(clusters))
    hexcolor = []
    for c in rgba:
      hexcolor.append(mpl.colors.rgb2hex(c))

    return hexcolor, cmap
      
def cluster_scheme(mol, hexcolor, clusters):
  r0 = mol[0]['resid']
  c0 = hexcolor[0]
  clust_scheme = []
  select = '@'
  i = 1
  j = 0
  for at in mol:
      r = at['resid']
      if r == r0:
          select += str(i) + ','
          i += 1
      else:
          clust_scheme.append([c0, select[:-1]])
          select = '@' + str(i) + ','
          r0 = r
          j +=1
          i +=1
          c0 = hexcolor[j]
          l0 = clusters[j]
  clust_scheme.append([c0, select[:-1]])

  return clust_scheme
  

def view_pdb_ngl(pdb, capsid, labels, rwb_scale=False):
    import biotite.structure.io as strucio
    strucio.save_structure(pdb + '_capsid.pdb', capsid, hybrid36=True)

    mol = open_pdb(pdb)
    hexcolor, cmap = clusters_colormap_hexcolor(labels, rwb_scale)
    clust_scheme = cluster_scheme(mol, hexcolor, labels)

    import nglview as ngl
    color_scheme = ngl.color._ColorScheme(clust_scheme, label="scheme_regions")
    view = ngl.show_structure_file(pdb + '_capsid.pdb', gui=False, default_representation=False)
    view._remote_call("setSize", target='Widget', args=['800px', '800px'])
    view.add_representation("spacefill", color=color_scheme)


    if rwb_scale:
        print('Each atom in this structure is colored according to the clustering quality score of its residue.')
        import matplotlib.colorbar as colorbar
        import matplotlib.pyplot as plt
        fig, ax = plt.subplots(figsize=(10,0.5))
        cb = colorbar.ColorbarBase(ax, orientation='horizontal',
                                   cmap=cmap, norm=plt.Normalize(np.min(labels), np.max(labels)))
        plt.show()
        code = """
                    var $text = $("<div></div>")
                                .css("position", "absolute")
                                .css("top", "3%")
                                .css("left", "30%")
                                .css("padding", "2px 5px 2px 5px")
                                .css("opacity", "1.0")
                                .css("font-size", "30px")
                                .css("color", "black")
                                .appendTo(this.$container);

                    $text.text("{0} residue cluster scores")
                    """
        view._execute_js_code(code.format(pdb))
    else:
        print('Each atom in this structure has the same color as other atoms in the same cluster.')
        code = """
            var $text = $("<div></div>")
                        .css("position", "absolute")
                        .css("top", "3%")
                        .css("left", "40%")
                        .css("padding", "2px 5px 2px 5px")
                        .css("opacity", "1.0")
                        .css("font-size", "30px")
                        .css("color", "black")
                        .appendTo(this.$container);

            $text.text("{0} ({1} clusters)")
            """
        n_clusters = str(int(np.max(labels) + 1))
        view._execute_js_code(code.format(pdb, n_clusters))


    view.center()
    print('rendering')
    view.render_image()

    return view

def createCapsidView(pdb, capsid=None):

    if capsid is None:
        print('No capsid structure provided, getting capsid.')
        from pyCapsid.PDB import getCapsid
        capsid, _, _, _, _, _, _ = getCapsid(pdb)

    import biotite.structure.io as strucio
    strucio.save_structure(pdb + '_capsid.pdb', capsid, hybrid36=True)

    import nglview as ngl
    view = ngl.NGLWidget()
    from nglview.adaptor import FileStructure

    view.add_component(FileStructure(pdb + '_capsid.pdb'))
    #view.clear_representations()

    return view

def showColorBar(labels):
    print('Each atom in this structure is colored according to the clustering quality score of its residue.')
    import matplotlib.colorbar as colorbar
    import matplotlib.pyplot as plt
    hexcolor, cmap = clusters_colormap_hexcolor(labels, rwb_scale=True)
    fig, ax = plt.subplots(figsize=(10, 0.5))
    cb = colorbar.ColorbarBase(ax, orientation='horizontal',
                               cmap=cmap, norm=plt.Normalize(np.min(labels), np.max(labels)))
    plt.show()

def createClusterRepresentation(pdb, labels, view, rwb_scale=False, rep_type='cartoon'):

    mol = open_pdb(pdb)
    hexcolor, cmap = clusters_colormap_hexcolor(labels, rwb_scale)
    clust_scheme = cluster_scheme(mol, hexcolor, labels)

    n_components = view.n_components
    for i in range(n_components):
        view.clear_representations(component=i)

    import nglview as ngl
    color_scheme = ngl.color._ColorScheme(clust_scheme, label="scheme_regions")
    view._remote_call("setSize", target='Widget', args=['800px', '800px'])
    view.add_representation(rep_type, color=color_scheme)

    if rwb_scale:
        print('Each atom in this structure is colored according to the clustering quality score of its residue.')
        import matplotlib.colorbar as colorbar
        import matplotlib.pyplot as plt
        fig, ax = plt.subplots(figsize=(10, 0.5))
        cb = colorbar.ColorbarBase(ax, orientation='horizontal',
                                   cmap=cmap, norm=plt.Normalize(np.min(labels), np.max(labels)))
        plt.show()
        code = """
                        var $text = $("<div></div>")
                                    .css("position", "absolute")
                                    .css("top", "3%")
                                    .css("left", "30%")
                                    .css("padding", "2px 5px 2px 5px")
                                    .css("opacity", "1.0")
                                    .css("font-size", "30px")
                                    .css("color", "black")
                                    .appendTo(this.$container);

                        $text.text("{0} residue cluster scores")
                        """
        view._execute_js_code(code.format(pdb))
    else:
        print('Each atom in this structure has the same color as other atoms in the same cluster.')
        code = """
                var $text = $("<div></div>")
                            .css("position", "absolute")
                            .css("top", "3%")
                            .css("left", "40%")
                            .css("padding", "2px 5px 2px 5px")
                            .css("opacity", "1.0")
                            .css("font-size", "30px")
                            .css("color", "black")
                            .appendTo(this.$container);

                $text.text("{0} ({1} clusters)")
                """
        n_clusters = str(int(np.max(labels) + 1))
        view._execute_js_code(code.format(pdb, n_clusters))

    view.center()
    print('rendering')
    #view.render_image()


def chimeraxReportScriptVis(report_path, remote=True, chimerax_path=None, **kwargs):
    """Launches ChimeraX and runs a script that visualizes the results.

    :param labels:
    :param pdb:
    :param remote:
    :param chimerax_path:
    :param pdb_path:
    :param save_path:
    :param script_path:
    :param labels_path:
    :return:
    """

    import os
    import platform

    if 'offscreen' in kwargs:
        offscreen = kwargs['offscreen']
    else:
        offscreen = False

    if chimerax_path is None:
        print('No chimerax path specified, checking in default locations for your OS')
        if platform.system() == 'Linux':
            chimerax_path = 'usr/bin/chimerax'
        elif platform.system() == 'Windows':
            chimerax_path = 'C:\\Program Files\\ChimeraX\\bin\\ChimeraX.exe'
        elif platform.system() == 'Darwin':
            chimerax_path = '/Applications/ChimeraX'
        else:
            print('No chimerax path is given and cannot check default locations')
            chimerax_path = ''

    # get path to chimerax script
    import os

    script_path = f'{report_path}/chimerax/chimerax_script_colab.py'

    chimerax_exe = chimerax_path  # + '\\ChimeraX.exe'
    if platform.system() == 'Windows':
        cmd_string = f'""{chimerax_exe}" --script "{script_path}""'
    elif platform.system() == 'Linux':
        cmd_string = f'{chimerax_exe} --script "{script_path}"'
    else:
        cmd_string = f'{chimerax_exe} --script "{script_path}"'

    if offscreen:
        cmd_string = cmd_string + ' --offscreen'
    print(cmd_string)
    os.system(cmd_string)